fixed_models:-
  fixed_models_data(Model, AllArrows, AllTraces, AllStates, Arrows, Traces, Links, StatesTrace),
  all_arrows(Model) := AllArrows, % for arrows, from 1 to 4
  all_traces(Model) := AllTraces, % for trace,  from 1 to 4
  init_fixed_events(Model, 0, 0, AllArrows, Arrows),
  init_fixed_events(Model, 0, 4, AllTraces, Traces),
  init_fixed_Links(Model, Links),
  init_fixed_States(Model, AllStates),
  init_fixed_Trace(Model, StatesTrace),
  fail.
fixed_models.

init_fixed_States(Model, AllStates):-
  for(I, 0, AllStates-1),
    for(J, 0, 3),
      states(32*Model+4*I+J) := 5,
      fail. 
init_fixed_States(Model, AllStates).

init_fixed_events(Model, Event, Start, AllEvents, [P_Event | List] ):-
  init_one_fixed_event(Model, Event, Start, P_Event, 0),
  init_fixed_events(Model, Event+1, Start, AllEvents, List).
init_fixed_events(Model, Event, Start, AllEvents, []).

init_one_fixed_event(Model, Event, Start, event(p( Atom, Type, Value, Negation)|Tail), Count):-
  events(80*Model+10*(Event+Start)+Count) := 6*G_K*2*Atom +G_K*2*Type + 2*Value + Negation,
  init_one_fixed_event(Model, Event, Start, event(|Tail), Count+1).
init_one_fixed_event(Model, Event, Start, event(), Count):-
  events(80*Model+10*(Event+Start)+Count) := max_integer.

init_fixed_Trace(Model, [trace(State, Event, Trace) | List] ):-
  states(32*Model+4*State+Event):= Trace,
  init_fixed_Trace(Model, List).
init_fixed_Trace(Model, []).

init_fixed_Links(Model, [link(State, State2, Event) | List] ):-
  links(64*Model+8*State+State2) := links(64*Model+8*State+State2)\/2**Event,
  init_fixed_Links(Model, List).
init_fixed_Links(Model, []).

%  fixed_models_data(Model, Arrows1-4, Traces1-4, AllStates, Arrows, Traces, Links, StatesTrace),
% event(p( Atom, Type, Value, Negation)|Tail)
% link(State, State2, Event)
% trace(State, Event, TraceVol)

fixed_models_data(0, 1, 1, 3,
 [event()], % true
 [event( p(0, 1, 2, 0))],
 [link(0, 1, 0), link(1, 2, 0), link(2, 0, 0)],
 [trace(2, 0, 2)]).

fixed_models_data(1, 2, 2, 8,
 [event( p(0, 0, 1, 0), p(0, 3, 0, 0)), event( p(0, 0, 2, 0), p(0, 3, 0, 0))],
 [event( p(0, 1, 1, 0), p(0, 3, 0, 0)), event( p(0, 1, 2, 0), p(0, 3, 0, 0))],
 [link(0, 1, 1), link(1, 2, 1), link(2, 3, 1), link(3, 4, 1), link(4, 5, 1), link(5, 6, 1), link(6, 7, 1),
  link(1, 0, 0), link(2, 1, 0), link(3, 2, 0), link(4, 3, 0), link(5, 4, 0), link(6, 5, 0), link(7, 6, 0)],
 [trace(0, 0, 2), trace(7, 1, 2)]).

fixed_models_data(2, 2, 2, 8,
 [event( p(0, 0, 1, 0), p(0, 3, 1, 0)), event( p(0, 0, 2, 0), p(0, 3, 1, 0))],
 [event( p(0, 1, 1, 0), p(0, 3, 1, 0)), event( p(0, 1, 2, 0), p(0, 3, 1, 0))],
 [link(0, 1, 1), link(1, 2, 1), link(2, 3, 1), link(3, 4, 1), link(4, 5, 1), link(5, 6, 1), link(6, 7, 1),
  link(1, 0, 0), link(2, 1, 0), link(3, 2, 0), link(4, 3, 0), link(5, 4, 0), link(6, 5, 0), link(7, 6, 0)],
 [trace(0, 0, 2), trace(7, 1, 2)]).

fixed_models_data(3, 1, 2, 2,
 [event( p(0, 0, 1, 0), p(0, 3, 2, 0))], % true
 [event( p(0, 1, 1, 0), p(7, 4, 1, 0), p(0, 3, 2, 0)), event( p(0, 1, 1, 0), p(7, 4, 1, 1), p(0, 3, 2, 0))],
 [link(0, 1, 0), link(1, 0, 0)],
 [trace(0, 1, 2), trace(1, 0, 2)]).

init_board:-
  for(I, 0, 7),
    init_squre(0, I, I+1),
    init_squre(1, I, 6+14),
    fail.
init_board:-
  for(I, 0, 2),
    init_squre(0, 7-I, I+1),
    fail.
init_board:-
  init_squre(0, 0, 1+14),
  init_squre(0, 4, 5+14),
  init_squre(0, 7, 1+14).

init_squre(Line, Column, Piece):-
    board(8*Line + Column):= Piece,
    board(56-8*Line + Column):= Piece+7.

make_chess_squre(X, I, J):-
  (board(8*J+I) =:= 7 ->
    color_text_back(_, rgb(255,255,0))
  else ((I+J) mod 2 =:=0 ->
    color_text_back(_, rgb(0,255,255))
  else
    color_text_back(_, rgb(240,240,240))
  )),
  (board(8*J+I) mod 14 > 6 ->
    color_text(_, rgb(0,0,0))
  else
    color_text(_, rgb(255,0,0))
  ),
  Piece := board(8*J+I) mod 7,
  (J =:= 9, active_previous(3) =:= 0 -> Piece := 0),
  face(Piece, Face),
  (I=:=0 -> 
    text_out(Face, pos(G_Zoom_Font*X, G_Zoom*(925-25*J)))
  else
    text_out(Face)
  ).

show_grid(I, J):-
  pen(3, rgb(0, 0, 0)),
  font(G_Zoom_Font*8, G_Zoom_Font*22, "Courier"), 
  X := G_Zoom*(300 - 35*I),
  Y := G_Zoom*(400 - 35*J),
  R := 70*G_Zoom,
  Step := 20*G_Zoom,
  color_text(_, rgb(0,0,255)),
  horizontal_grid(X, Y, I, J, R),
  vertical_grid(X, Y, I, J, R),
  color_text(_, rgb(0,0,0)),
  for(B, 0, J-1),
    for(A, 0, I-1),
      (G_Sel =:= 4 ->
        Object := board(8*(7-B)+A)
      else
        Object := board(64+A)
      ),
      Object > 0,
      (Object =:= 7 ->
        color_text(_, rgb(0,255,0)),
        names(14, _, _, _, Name),
        text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
        fail
      else (Object mod 14 > 6 ->
        names(6, _, _, _, Name),
        color_text(_, rgb(0,0,0))
      else
        names(7, _, _, _, Name),
        color_text(_, rgb(255,0,0))
      )),
      text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
      color_text(_, rgb(0,0,255)),
      Piece := 7+ Object mod 7,
      names(Piece, _, _, _, Name2),
      text_out(Name2, pos(2+X+A*R, 3+Step+Y+B*R)),
      Object > 14,
      color_text(_, rgb(0,0,0)),
      names(15, _, _, _, Name3),
      text_out(Name3, pos(2+X+A*R, 3+2*Step+Y+B*R)),
      fail.
show_grid(I, J):-
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"). 

horizontal_grid(X, Y, I, J, R):-
  for(B, 0, J),
    line(X, Y+B*R, X+I*R, Y+B*R),
    B < J, J > 1,
    text_out(J-B-1, pos(X-G_Zoom*17, G_Zoom*25+Y+B*R)),
    fail.
horizontal_grid(X, Y, I, J, R).

vertical_grid(X, Y, I, J, R):-
  for(A, 0, I),
    line(X+A*R, Y, X+A*R, Y+J*R),
    A < I, I > 1,
    text_out(A, pos(X+G_Zoom*30+A*R, Y+G_Zoom*5+J*R)),
    fail.
vertical_grid(X, Y, I, J, R).

observable(Atom) :-
  (Atom > 5, Atom < 16 ->
    (G_Observed > 0 ->
      Atom =:= G_Observed
    else
      Squre := G_ObserveNow,
      Squre > 0,
      (Atom < 8 -> % colour
        (Squre mod 14 > 6 -> Atom=:=6, Squre=\=7 else Atom=:=7)
      else (Atom < 14 -> % piece
        Atom - 7 =:= Squre mod 7
      else
        (Atom=:=14, Squre =:=7; Atom=:=15, Squre > 14)
      ))
    )
  ).

up:- % m0=2 & a0=a & m3=0
  active_previous(0)=:=2,
  buf_actions(G_Time mod 50)=:=1,
  active_previous(3)=:=0.

down :- % m0=2 & a0=a & m3=1
  active_previous(0)=:=2,
  buf_actions(G_Time mod 50)=:=1,
  active_previous(3)=:=1.

here(I, J):-
  I := active_previous(1),
  J := active_previous(2).

somewhere(I, J):-
  repeat,
  I := random(8),
  J := random(8).

update_memory :- % up, here => copy(Lifted)
  up,
  here(I, J),
  board(8*J+I) := 7,
  fail.
update_memory :- % down, everywhere => del(Lifted)
  down,
  for(I, 0, 7),
    for(J, 0, 7),
      (board(8*J+I) =:= 7 -> board(8*J+I) := 0),
      fail.
update_memory :- % down, here => copy(mem5(0))
  down,
  here(I, J),
  board(8*J+I) := board(64+0),
  fail.
update_memory :- % down, !ob(Lifted), here => del(Unmoved)
  down,
  G_ObserveNow =\= 7,
  here(I, J),
  board(8*J+I) := board(8*J+I) mod 14,
  fail.
update_memory :- % down, !ob(Lifted), somewhere, Black in mem4(there) => Temp := mem4(there), del(all)
  down,
  G_ObserveNow =\= 7,
  somewhere(I, J),
  board(8*J+I) mod 14 > 6,
  board(8*J+I) =\= 7,
  cut(0.5),
  G_Temp := board(8*J+I),
  board(8*J+I) := 0,
  fail.
update_memory :- % down, !ob(Lifted), somewhere, !(Black in mem4(there)) => mem4(there) := Temp, del(Unmoved)
  down,
  G_ObserveNow =\= 7,
  somewhere(I, J),
  not(board(8*J+I) mod 14 > 6, board(8*J+I) =\= 7),
  cut(0.5),
  board(8*J+I) := G_Temp,
  board(8*J+I) := board(8*J+I) mod 14,
  fail.
update_memory :- %  => ob(now) := mem4(here)
  I := active(1),
  J := active(2),
  G_ObserveNow := board(8*J+I),
  fail.
/*----------------*/
update_memory :- %  => mem5(here) := ob(now)
  I := active(3),
  board(64+I) := G_ObserveNow,
  fail.
update_memory :- % down, where=1 => copy(Lifted)
  down,
  I := 1,
  board(64+I) := 7,
  fail.
update_memory.

show_memory_rules:-
  I := 0,
  memory_rules(Number, G_Sel, Text1, Text2),
  I := I + 1,
  color_text(_, rgb(255,0,0)),
  text_out(Number, pos(G_Zoom*600, G_Zoom*(30+I*40))),
  color_text(_, rgb(0,0,0)),
  text_out(Text1),
  color_text(_, rgb(0,0,255)),
  text_out(Text2).


memory_rules( "1. ", 4, "  up, here =>", " copy(Lifted)").
memory_rules( "2. ", 4, "down, everywhere =>", " del(Lifted)").
memory_rules( "3. ", 4, "down, here =>", " copy(mem5(0))").
memory_rules( "4. ", 4, "move, here =>", " del(Unmoved)").
memory_rules( "", 4, "", "").
memory_rules( "5. ", 4, "move, somewhere, Black is there =>", "").
memory_rules( "", 4, "=>", " Temp := mem4(there), del(all)").
memory_rules( "", 4, "", "").
memory_rules( "6. ", 4, "move, somewhere, !(Black is there) =>", "").
memory_rules( "", 4, "=>", " mem4(there) := Temp, del(Unmoved)").
memory_rules( "", 4, "", "").
memory_rules( "7. ", 4, "=>", " ob(now) := mem4(here)").

memory_rules( "", 5, "", "").
memory_rules( "", 5, "", "").
memory_rules( "", 5, "", "").
memory_rules( "", 5, "", "").
memory_rules( "1. ", 5, "=>", " mem5(here) := ob(now)").
memory_rules( "2. ", 5, "down, where=1 =>", " copy(Lifted)").

memory_rules( "", _, "", "").
memory_rules( "", 4, "up   =", " m0=2 & a0=a & m3=0").
memory_rules( "", _, "down =", " m0=2 & a0=a & m3=1").
memory_rules( "", 4, "move =", " down, !ob(Lifted)").

make_colors(4):-
  color_text(_, rgb(255,0,0)).
make_colors(6):-
  color_text(_, rgb(0,0,0)),
  color_text_back(_, rgb(0,255,255)).
make_colors(7):-
  color_text(_, rgb(255,255,255)).
make_colors(8):-
  color_text(_, rgb(0,0,255)),
  color_text_back(_, rgb(255,255,0)).
make_colors(14):-
  color_text(_, rgb(0,0,0)),
  color_text_back(_, rgb(0,255,0)).


face(0, "  ").
face(1, " r").
face(2, " k").
face(3, " b").
face(4, " Q").
face(5, " K").
face(6, "p ").

names(0, "0", "            1, 2, 3", "", "").
names(1, "1", "           Horizontal", "", "").
names(2, "2", "            Vertical", "", "").
names(3, "3", "Lifting Up and Placing Down", "", "").
names(4, "M", "", "", "").
names(5, "M", "Memory - One Dimension (Lift - Place)", "", "").
names(6, "B", "           Black Colour", "Black", "black").
names(7, "W", "          White Colour", "White", "white").
names(8, "r", "             Rook", "Rook", "rook").
names(9, "k", "             Knight", "Knight", "knigh").
names(10, "b", "             Bishop", "Bishop", "bisho").
names(11, "Q", "             Queen", "Queen", "queen").
names(12, "K", "             King", "King", "king").
names(13, "p", "             Pawn", "Pawn", "pawn").
names(14, "L", "  A Piece is Lifted from Here", "Lifted", " lift").
names(15, "U", "         Unmoved Piece", "Unmoved", "unmov").

